<?php
/**
 * zogan - PHP template engine for XML
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * @author Haruki Setoyama <haruki@planewave.org>
 * @copyright Copyright &copy; 2003-2004, Haruki SETOYAMA <haruki@planewave.org>
 * @license http://www.opensource.org/licenses/lgpl-license.php LGPL
 * @version 0.2 $Id: zogan.php,v 1.16 2004/10/19 03:55:40 haruki Exp $
 * @link http://phpoot.sourceforge.jp/zogan/
 * @package zogan
 */
/**
 * Requires PEAR
 */
require_once 'PEAR.php';

// Constants
/**
 * mark for compiled template that mean ZOGAN library has been included
 */
define('ZOGAN', true);
/**
 * directory where ZOGAN library is in
 */
@define('ZOGAN_DIR', dirname(__FILE__).'/zogan');
/**
 * directory where fuctions for template are in
 */
@define('ZOGAN_FUNCTION_DIR', ZOGAN_DIR.'/functions/');
/**
 * extention added to compiled template
 */
@define('ZOGAN_EXT', '.php');
/**
 * parser to compiled template
 */
@define('ZOGAN_PARSER', 'zogan.htmlparser.php');
//@define('ZOGAN_PARSER', 'zogan.parser.php');

// error codes
define('ZOGAN_ERR_INTERNAL',    1);
define('ZOGAN_ERR_TEMPLATE',    2);
define('ZOGAN_ERR_COMPILE_DIR', 4);
define('ZOGAN_WARN_TEMPLATE',   8);

// options
define('ZOGAN_OPTION_FORCE_COMPILE',    1);
define('ZOGAN_OPTION_CHECK_TIMESTAMP',  2);
define('ZOGAN_OPTION_DEBUG', 4);

/**
 * ZOGAN Main Class.
 *
 * usage:
 * <pre>
 *  require_once 'XML/Template/zogan.php';
 *  $tmpl = new zogan;
 *  $tmpl->setTemplateFile('./helloworld.html');
 *  $tmpl->setCompileDir('./DirForCompile');
 *  $tmpl->display($model_data);
 * </pre>
 *
 * For more information, see http://phpoot.sourceforge.jp/zogan/
 * or http://www.planewave.org/zogan/ .
 *
 * @access public
 * @package zogan
 * @subpackage interface
 */
class Zogan extends PEAR {

    /**
     * path to directory where compiled files will be stored
     * @var string
     * @access private
     */
    var $_compiled_dir;
    /**
     * path of compiled file
     * @var string
     * @access private
     */
    var $_compiled_file;
    /**
     * template file abstracton
     * @var zogan_templatefile
     * @access private
     */
    var $_template_file;
    /**
     * data to be embedded in the template
     * @access private
     */
    var $_data;
    /**
     * debug option
     * @access private
     */
    var $_debug = false;
    /**
     * force_compile option
     * @access private
     */
    var $_force_compile = false;
    /**
     * check_timestamp option
     * @access private
     */
    var $_check_timestamp = true;
    /**
     * functions to apply for model data
     * @access private
     */
    var $_functions = array();
    /**
     * options for functions to apply for model data
     * @access private
     */
    var $_f_options = array();
    /**
     * keys without XML entity escape
     * @access private
     */
    var $_no_escape = array();
    /**
     * pi name not to be removed
     * @access private
     */
    var $_pi_kept = array('xml');
    /**
     * names and methods of object to be treated as data
     * @access private
     */
    var $_object_for_str = array('zogan' => 'fetch');
    /**
     * error objects
     * @access private
     */
    var $_error_class = 'Zogan_Error';

    /**
     * constructor
     */
    function Zogan()
    {
        // do nothig now
    }

    /**
     * sets the path for the template file or the template abstraction
     *
     * @param stirng|Zogan_Template $template
     * @return true|PEARError
     */
    function setTemplateFile($template)
    {
        if (is_string($template)) {
            require_once ZOGAN_DIR.'/zogan.templatefile.local.php';
            $this->_template_file = new zogan_templatefile_local($template);
        }
        elseif (is_a($template, 'zogan_templatefile')) {
            $this->_template_file = $template;
        }
        else {
            return $this->raiseError(
                "Invalid template file specified at setTemplateFile(). ",
                ZOGAN_ERR_TEMPLATE);
        }
        return true;
    }

    /**
     * sets the path for the directory where compiled files will be stored
     *
     * @param stirng $dirname directory name
     * @return true
     */
    function setCompileDir($dirname)
    {
        $this->_compiled_dir = $dirname;
        return true;
    }

    /**
     * sets the variable names to be un-escaped
     *
     * @param array
     */
    function setNoEscapeVars($vars)
    {
        $this->_no_escape = array();
        foreach ($vars as $var) {
            $this->_no_escape[$var] = 1;
        }
    }

    /**
     * adds the variable name to be un-escaped
     *
     * @param string
     */
    function addNoEscapeVar($var)
    {
        $this->_no_escape[$var] = 1;
    }

    /**
     * sets function resouces to modify model data
     *
     * @param array $funcs
     */
    function setFunctions($funcs)
    {
        $this->_functions = $funcs;
    }

    /**
     * adds a function resouce to modify model data
     *
     * @param string $name
     * @param resource $res
     */
    function addFunction($name, $res)
    {
        $this->_functions[$name] = $res;
    }

    /**
     * sets a function specific option
     *
     * @param string $name
     * @param mixed $option
     */
    function setFunctionOption($name, $option)
    {
        $this->_f_options[$name] = $option;
    }

    /**
     * adds a pi target to be kept
     *
     * @param string $target
     */
    function addPiKept($target)
    {
        if (!in_array($target, $this->_pi_kept)) {
            $this->_pi_kept[] = $target;
        }
    }
    
    /**
     * adds a pi target to be kept
     *
     * @param string $class classname
     * @param string $method method name to get data
     */
    function addClassForData($class, $method)
    {
        $this->_object_for_str[$class] = $method;
    }

    /**
     * sets option
     *
     * @param string $key
     * @param bool $val
     * @return bool
     */
    function setOption($key, $val)
    {
        switch($key){
            case 'force_compile':
            case ZOGAN_OPTION_FORCE_COMPILE:
                $this->_force_compile = (bool)$val;
                return true;
                break;
            case 'check_timestamp':
            case ZOGAN_OPTION_CHECK_TIMESTAMP:
                $this->_check_timestamp = (bool)$val;
                return true;
                break;
            case 'debug':
            case ZOGAN_OPTION_DEBUG:
                $this->_debug = (bool)$val;
                return true;
                break;
            default:
                return false;
        } // switch
    }

    /**
     * sets data
     */
    function setData($data)
    {
        $this->_data = $data;
    }

    /**
     * displays the template results
     *
     * @param array $data   model data
     * @return bool|PEARError
     */
    function display($data =null)
    {
        $ret = $this->_check();
        if ($ret !== true) return $ret;
    
        if ($this->_need_compile() !== false) {
            // compile
            $err = $this->_compile();
            if ($err !== true) return $err;
        }

        if ($this->_debug
        || $this->_default_error_mode & PEAR_ERROR_PRINT) {
            $this->_quiet = false;
        } else {
            $this->_quiet = true;
        }

        if ($data === null) {
            $_val0 = $this->_data;
        } else {
            $_val0 = $data;
        }
        include $this->_compiled_file;
        return true;
    }

    /**
     * checks if the template needs to be compiled
     *
     * @access private
     * @return bool
     */
    function _need_compile()
    {
        if ($this->_force_compile != false || $this->_debug != false) {
            return true;
        }
        if (file_exists($this->_compiled_file) != true) {
            return true;
        }
        if ($this->_check_timestamp != false
        && $this->_template_file->getUpdateTime() > filemtime($this->_compiled_file)) {
            return true;
        }
        return false;
    }

    /**
     * gets the template results as string
     *
     * @param sting $output     variable for output
     * @param array $data       model data
     * @return bool|PEARError
     */
    function expand(&$output, $data =null)
    {
        ob_start();
        $ret = $this->display($data);
        $output = ob_get_contents();
        ob_end_clean();
        return $ret;
    }

    /**
     * returns the template results as string
     *
     * @param array $data       model data
     * @return string
     */
    function fetch($data =null)
    {
        ob_start();
        $this->display($data);
        $output = ob_get_contents();
        ob_end_clean();
        return $output;
    }

    /**
     * compiles the template file
     *
     * @return true|PEARError
     */
    function compile()
    {
        $ret = $this->_check();
        if ($ret !== true) return $ret;
        
        return $this->_compile();
    }
        
    /**
     * compiles the template file
     *
     * @return true|PEARError
     * @access private
     */
    function _compile()
    {
        require_once ZOGAN_DIR.'/localfile.php';

        $file =& LocalFile::create($this->_compiled_file);
        $file->_default_error_mode =& $this->_default_error_mode;
        $file->_default_error_options =& $this->_default_error_optons;

        $lock = $file->lock();
        if ($lock !== true) { return $lock; }

        $parsed = $this->_parse();
        if (PEAR::isError($parsed)) {
            $file->unlock();
            return $parsed;
        }
        $parsed = "<?php if(! defined('ZOGAN')) exit(); ?>\n"
                    . $parsed; // <?

        return $file->write($parsed);
    }

    /**
     * removes the compiled template file
     *
     * @return true|PEARError
     */
    function removeCompiled()
    {
        $ret = $this->_check();
        if ($ret !== true) return $ret;

        if (file_exists($this->_compiled_file)) {
            if (! unlink($this->_compiled_file)) {
                return $this->raiseError(
                        "Failed to remove the compiled file '$this->_compiled_file': $php_errormsg. ",
                        ZOGAN_ERR_COMPILE_DIR);
            }
        }
        return true;
    }

    /**
     * parses the template file
     *
     * @access private
     * @return string|PEARError parsed PHP code
     */
    function _parse()
    {
        $template = $this->_template_file->getTemplate();
        if (PEAR::isError($template)) return $template;

        require_once ZOGAN_DIR . '/' . ZOGAN_PARSER;
        $parser = new zogan_parser($this, $this->_template_file->getEncoding());

        return $parser->build($template);
    }
    
    /**
     * checks if template and directory for compiled file are set 
     *
     * @access private
     * @return true|PEARError
     */
    function _check()
    {
        if (! isset($this->_template_file)) {
            return $this->raiseError(
                    "Template file is not set. ",
                    ZOGAN_ERR_TEMPLATE);
        }
        if (!isset($this->_compiled_dir) || $this->_compiled_dir == '') {
            return $this->raiseError(
                    "Directory for compiled file is not set. ",
                    ZOGAN_ERR_COMPILE_DIR);
        }
        $this->_compiled_file
            = $this->_compiled_dir . '/'
            . $this->_template_file->getBasename() . ZOGAN_EXT;
        return true;
    }

    /**
     * escapes string with XML predefined entity
     * @return string
     * @access private
     */
    function _escape($val)
    {
        return htmlspecialchars($val);
        /*return str_replace(array('<','>','&','\'','"')
                            , array('&lt;', '&gt;', '&amp;', '&apos;', '&quot;')
                            ,$val);*/
    }

    // functions with "_c_" prefix are used in compiled template file.

    /**
     * get value from $val with $var as the name of scaler variable
     * @return mixed
     * @access private
     */
    function _c_val($val, $var)
    {
        if (is_array($val) && isset($val[$var])) {
            $ret = $val[$var];
        }
        elseif (is_object($val) && substr($var, 0, 1) != '_') {
            if (is_callable(array($val, $var))) {
                $ret = $val->$var();
            } else {
                $ret = $val->$var;
            }
        }
        else {
            $ret = null;
        }

        return $ret;
    }

    /**
     * get value from $val with $vars as the name of array
     * @return mixed
     * @access private
     */
    function _c_val_array($val, $vars)
    {
        $var = array_shift($vars);
        $ret = $this->_c_val($val, $var);
        if (count($vars) > 0) {
            return $this->_c_val_array($ret, $vars);
        } else {
            return $ret;
        }
    }

    /**
     * returns array of data for nested variabe tag
     * not hash!
     * @return array
     * @access private
     */
    function _c_each($target)
    {
        if (is_array($target) && isset($target[0])) {
            return $target;
        }
        return array($target);
    }

    /**
     * change scalar and array data to hash
     * @access private
     */
    function _c_hash(&$scalar, $var)
    {
        if (is_scalar($scalar)) {
            $arr[$var] = $scalar;
            $scalar = $arr;
        }
        elseif (is_array($scalar) && isset($scalar[0])) {
            $temp = array();
            foreach($scalar as $val) {
                $temp[] = array($var => $val);
            }
            $scalar = $temp;
        }
    }

    /**
     * check if $val is printable, that is string or numeric
     * @access private
     */
    function _c_is_string($val)
    {
        return (is_string($val) || is_numeric($val));
    }

    /**
     * echos string from data for attribute
     * @access private
     */
    function _c_attr($attr, $value, $default)
    {
        if (is_object($value)) {
            $value = $this->_c_obj_to_str($value);
        }
        if ($value === true) {
            echo ' ' . $attr . '="' . htmlspecialchars($default) . '"';
        }
        elseif ($this->_c_is_string($value)) {
            echo ' ' . $attr . '="' . htmlspecialchars($value) . '"';
        }
        else {
            echo '';
        }
    }

    /**
     * echos string from data for content
     * @access private
     */
    function _c_content($val, $variable, $escape =true)
    {
        if (is_object($val)) {
            $val = $this->_c_obj_to_str($val);
            $escape = false;
        }

        if ($this->_c_is_string($val)) {
            if ($escape != true || isset($this->_no_escape[$variable])) {
                echo $val;
            } else {
                echo $this->_escape($val);
            }
        }
        else {
            echo (! $this->_quiet) ? "<!-- Data type of '$variable' is invalid. -->" : '';
        }
    }

    /**
     * gets string expression of registerd object
     * @access private
     */
    function _c_obj_to_str($val)
    {
        foreach ($this->_object_for_str as $classname => $method) {
            if (is_a($val, $classname)) {
                return $val->$method();
            }
        }
        return $val;
    }

    /**
     * change basename of template for file inclution
     * @return bool
     * @access private
     */
    function _c_change_basename($basename)
    {
        $this->_template_file->changeBasename($basename);
        //$this->_set_compiled_file();
        $this->_compiled_file
            = $this->_compiled_dir . '/'
            . $this->_template_file->getBasename() . ZOGAN_EXT;
        return true;
    }

    /**
     * apply function to model data
     * @return mixed
     * @access private
     */
    function _c_func($val, $func_name, $param)
    {
        $op = (isset($this->_f_options[$func_name]))
            ? $this->_f_options[$func_name] : null ;
        if (isset($this->_functions[$func_name])) {
            if (is_callable($this->_functions[$func_name])) {
                return call_user_func($this->_functions[$func_name], $val, $param, $op);
            }
        }

        if (file_exists(ZOGAN_FUNCTION_DIR.$func_name.'.php')) {
            require_once ZOGAN_FUNCTION_DIR.$func_name.'.php';
            $function = 'zogan_func_'.$func_name;
            return $function($val, $param, $op);
        }

        if (!$this->_c_is_string($val)) {
            return $val;
        }
        else {
            return $this->_escape($val);
        }
    }

    /**
     * Is the target allowed pi?
     * @return bool
     * @access private
     */
    function _c_pi($target)
    {
        return in_array($target, $this->_pi_kept);
    }
}

/**
 * ZOGAN_Error
 *
 * @package zogan
 * @subpackage interface
 */
class Zogan_Error extends PEAR_Error 
{
    function Zogan_Error($message = 'unknown error', $code = null,
                         $mode = null, $options = null, $userinfo = null)
    {
        $this->PEAR_Error("Zogan Error: $message\n", $code, $mode, $options, $userinfo);
    }
}

/**
 * ZOGAN template file abstraction
 *
 * @access public
 * @package zogan
 * @subpackage template
 * @abstract
 */
class zogan_templatefile extends PEAR {

    /**
     * constructor
     */
    function zogan_templatefile($file) {}

    /**
     * returns update time of template
     * @return int UNIX time
     */
    function getUpdateTime() {}

    /**
     * returns basename of template file
     * This name will be used for compiled file
     * @return string
     */
    function getBasename() {}

    /**
     * returns template string
     * @return string
     */
    function getTemplate() {}

    /**
     * change only basename of template
     * @return void
     */
    function changeBasename($file) {}

    /**
     * returns charactor encoding of template.
     * UTF-8 is recommended for XML template.
     * @return string
     */
    function getEncoding() {}
}

/**
 * function to clone object
 */
if ((int)substr(PHP_VERSION,0,1) >= 5) {
    /**
     * @access private
     */
    function zogan_clone_obj($obj)
    {
        return clone($obj);
    }
} else {
    /**
     * @access private
     */
    function zogan_clone_obj($obj)
    {
        return $obj;
    }
}

?>